#ifdef TARGET_DEFS_ONLY
#define NB_REGS 32

#define PTR_SIZE      4
#define LDOUBLE_SIZE  8
#define MAX_ALIGN     4
#define LDOUBLE_ALIGN 4

#define RC_INT   0x0001
#define RC_FLOAT 0x0001
#define RC_RV    0x0100
#define RC_RVH   0x0200

#define RC_IRET RC_RV
#define RC_LRET RC_RVH
#define RC_FRET RC_RV

enum {
  TREG_ZR = 0,
  TREG_SP,  TREG_FP,  TREG_R3,  TREG_R4,  TREG_R5,  TREG_R6,
  TREG_R7,  TREG_R8,  TREG_LR,  TREG_R10, TREG_RV,  TREG_RVH,
  TREG_R13, TREG_R14, TREG_R15, TREG_R16, TREG_R17, TREG_R18,
  TREG_R19, TREG_R20, TREG_R21, TREG_R22, TREG_R23, TREG_R24,
  TREG_R25, TREG_R26, TREG_R27, TREG_R28, TREG_R29, TREG_R30,
  TREG_R31
};

#define REG_IRET TREG_RV
#define REG_LRET TREG_RVH
#define REG_FRET TREG_RV

#define TARGET_BIG_ENDIAN

#ifndef __OR1K__
#define TARGET_ENDIAN_REVERSING
#endif

#define TARGET_NOCARRY

#define TCC_SOFTFLOAT
#else

#include "tcc.h"

ST_DATA const int reg_classes[NB_REGS] = {
    /* r0  */ 0,        /* r1  */ 0,        /* r2  */ 0,   /* r3  */ 0,
    /* r4  */ 0,        /* r5  */ 0,        /* r6  */ 0,   /* r7  */ 0,
    /* r8  */ 0,        /* r9  */ 0,        /* r10 */ 0,   /* r11 */ RC_RV,
    /* r12 */ RC_RVH,   /* r13 */ RC_INT,   /* r14 */ 0,   /* r15 */ RC_INT,
    /* r16 */ 0,        /* r17 */ RC_INT,   /* r18 */ 0,   /* r19 */ RC_INT,
    /* r20 */ 0,        /* r21 */ RC_INT,   /* r22 */ 0,   /* r23 */ RC_INT,
    /* r24 */ 0,        /* r25 */ RC_INT,   /* r26 */ 0,   /* r27 */ RC_INT,
    /* r28 */ 0,        /* r29 */ RC_INT,   /* r30 */ 0,   /* r31 */ RC_INT
};

static uint32_t func_sub_sp_offset;
static int func_ret_sub;

ST_FUNC void g(int c) {
  int ind1;
  if (nocode_wanted) return;

  ind1 = ind + 1;
  if ( ind1 > cur_text_section->data_allocated )
  	section_realloc( cur_text_section, ind1 );

  cur_text_section->data[ind] = c;
  ind = ind1;
}

ST_FUNC void o(unsigned int c) {
  g( c >> 24 );
  g( c >> 16 );
  g( c >> 8 );
  g( c );
}

uint32_t read32be( unsigned char *ptr ) {
  return (ptr[0] << 24) | (ptr[1] << 16) | (ptr[2] << 8) | ptr[3];
}

void write32be( unsigned char *ptr, uint32_t v ) {
  ptr[0] = v >> 24;
  ptr[1] = v >> 16;
  ptr[2] = v >> 8;
  ptr[3] = v;
}

int rel26dec( uint32_t code ) {
  return (code & 0x3ffffff) | ((code & 0x2000000)? 0xfc000000: 0);
}

uint32_t rel26enc( unsigned char jcode, int disp ) {
  return (jcode << 26) | ((disp >> 2) & 0x3ffffff);
}

void l_addi( int dst, int src, int imm ) {
  o((0x27 << 26) | (dst << 21) | (src << 16) | (imm & 0xffff)); // l.addi  dst, src, imm16
}

void l_ori( int dst, int src, uint32_t imm ) {
  o((0x2a << 26) | (dst << 21) | (src << 16) | (imm & 0xffff)); // l.ori dst, src, imm16
}

ST_FUNC void gsym_addr(int t, int a) {
  while ( t ) {
  	unsigned char *ptr = cur_text_section->data + t;

	uint32_t n = read32be( ptr );
	unsigned char cmd = n >> 26;
	int32_t disp = rel26dec(n);

        if ( cmd == 0 || cmd == 3 || cmd == 4 ) // l.j, l.bnf, l.bf
		write32be( ptr, rel26enc(cmd, a - t) );
	else
		printf("gsym 0x%x -> 0x%0x failed\n", t, ind);

	if ( !disp ) break;
	t = t + disp * 4;
  }
}

ST_FUNC int gjmp(int t) {
  int ind1 = ind;
  if (nocode_wanted) return t;
  o(rel26enc(0, t - ind1)); // l.j t
  o(0x15000000); // l.nop
  return ind1;
}

ST_FUNC void gjmp_addr(int a) {
  gjmp( a );
}

int fpdbl_type( int ft ) {
  if ( ft == VT_LDOUBLE ) return VT_DOUBLE;
  return ft;
}

ST_FUNC void load(int r, SValue *sv) {
  int v, fr, ft, fc, size = 0;
//  char unsign = 0;
  SValue v1;

  fr = sv->r;
  ft = sv->type.t;
  fc = sv->c.i;

  v = fr & VT_VALMASK;
//  printf( "load r=%d sv->r=0x%x\n", r, fr );

  if ( fr & VT_LVAL ) {
        if (v == VT_LLOCAL) {
            v1.type.t = VT_INT;
            v1.r = VT_LOCAL | VT_LVAL;
            v1.c.i = fc;
            load(r, &v1);
            fr = r;
	    v  = r;
        }
	if ((ft & VT_TYPE) == VT_BYTE) {
            size = 1;
        } else if ((ft & VT_TYPE) == (VT_BYTE | VT_UNSIGNED)) {
            size = 1;
//            unsign = 1;
        } else if ((ft & VT_TYPE) == VT_SHORT) {
            size = 2;
        } else if ((ft & VT_TYPE) == (VT_SHORT | VT_UNSIGNED)) {
            size = 2;
//            unsign = 1;
        } else if (fpdbl_type(ft & VT_BTYPE) == VT_DOUBLE) {
            size = 8;
        } else {
            size = 4;
        }

	if ( (fr & VT_VALMASK) < VT_CONST ) {
		if ( size == 1 )
			o( (0x23 << 26) | (r << 21) | (v << 16) );  // l.lbz r, 0(v)
		else if ( size == 2 )
			o( (0x25 << 26) | (r << 21) | (v << 16) );  // l.lhz r, 0(v)
		else if ( size == 4 )
			o( (0x21 << 26) | (r << 21) | (v << 16) );  // l.lwz r, 0(v)
		else if ( size == 8 )
			tcc_error("load long long not supported");
	}
  	else if ( fr & VT_SYM ) {
		greloc(cur_text_section, sv->sym, ind,     R_OR1K_HI_16_IN_INSN);
		greloc(cur_text_section, sv->sym, ind + 4, R_OR1K_LO_16_IN_INSN);
		o( (6 << 26) | (r << 21) ); // l.movhi r,    sym_hi
		l_ori( r, r, 0 );           // l.ori   r, r, sym_lo

		if ( size == 1 )
			o( (0x23 << 26) | (r << 21) | (r << 16) | (fc & 0xffff) );  // l.lbz r, fc(r)
		else if ( size == 2 )
			o( (0x25 << 26) | (r << 21) | (r << 16) | (fc & 0xffff) );  // l.lhz r, fc(r)
		else if ( size == 4 )
			o( (0x21 << 26) | (r << 21) | (r << 16) | (fc & 0xffff) );  // l.lwz r, fc(r)
		else if ( size == 8 )
			tcc_error("load long long not supported");
	}
	else {
		if ( size == 1 )
			o( (0x23 << 26) | (r << 21) | (2 << 16) | (fc & 0xffff) );  // l.lbz r, fc(fp)
		else if ( size == 2 )
			o( (0x25 << 26) | (r << 21) | (2 << 16) | (fc & 0xffff) );  // l.lhz r, fc(fp)
		else if ( size == 4 )
			o( (0x21 << 26) | (r << 21) | (2 << 16) | (fc & 0xffff) );  // l.lwz r, fc(fp)
		else if ( size == 8 )
			tcc_error("load long long not supported");
	}
  }
  else {
  	if ( v == VT_CONST ) {
		if ( fr & VT_SYM ) {
			greloc(cur_text_section, sv->sym, ind,     R_OR1K_HI_16_IN_INSN);
			greloc(cur_text_section, sv->sym, ind + 4, R_OR1K_LO_16_IN_INSN);
			o( (6 << 26) | (r << 21) ); // l.movhi r, const_hi
			l_ori( r, r, 0 );           // l.ori   r, r, const_lo
			if ( fc ) l_addi( r, r, fc );
		}
		else {
			if ( fc >> 16 ) {
				o( (6 << 26) | (r << 21) | ((fc >> 16) & 0xffff) ); // l.movhi r, const_hi
				l_ori(r, r, fc );                                   // l.ori   r, r, const_lo
			}
			else
				l_ori(r, 0, fc );                                   // l.ori   r, r0, const_lo
		}
	}
	else if ( v == VT_LOCAL )
		l_addi( r, 2, fc );

	else if ( v == VT_CMP ) {
#if OR1K_CMOV_ENABLED
		int r2 = get_reg(RC_INT);
		l_ori( r2, 0, 1 );                                // l.ori  r2, r0, 1
		if ( fc & 1 )
			o( (0x38 << 26) | (r2 << 21) | (r << 16) | 0xe ); // l.cmov r, r0, r2
		else
			o( (0x38 << 26) | (r << 21) | (r2 << 16) | 0xe ); // l.cmov r, r2, r0
#else
		o( (4 << 26) | 3 ); // l.bf pc+3
		fc &= 1;
		l_ori( r, 0, fc ^ 1 );  // l.ori  r, r0, 1
		l_ori( r, 0, fc     );  // l.ori  r, r0, 0
#endif
	}
	else if ( v == VT_JMP || v == VT_JMPI ) {
		int t = v & 1;
		o( 3 );              // l.j 3
		l_ori( r, 0, t );
		gsym( fc );
		l_ori( r, 0, t ^ 1 );
	}
	else if ( v != r ) // load value from other reg
		l_ori( r, v, 0 );

//	else
//		tcc_warning("load nop");
  }
}

ST_FUNC void store(int r, SValue *v) {
  int fr, bt, ft, fc, size;
 
  ft = v->type.t;
  fc = v->c.i;
  fr = v->r & VT_VALMASK;
  bt = ft & VT_BTYPE;

//  printf( "store 0x%x, fc=0x%x\n", fr, fc );

  if (bt == VT_SHORT)
	size = 2;
  else if (bt == VT_BYTE)
	size = 1;
  else if (fpdbl_type(bt) == VT_DOUBLE)
	size = 8;
  else
	size = 4;

  if ((v->r & VT_VALMASK) == VT_CONST) {
	int r2 = get_reg(RC_INT);

	if ( v->r & VT_SYM ) {
		greloc(cur_text_section, v->sym, ind,     R_OR1K_HI_16_IN_INSN);
		greloc(cur_text_section, v->sym, ind + 4, R_OR1K_LO_16_IN_INSN);
		o( (6 << 26) | (r2 << 21) );  // l.movhi r2, const_hi
		l_ori( r2, r2, 0 );           // l.ori   r2, r2, const_lo
		if ( fc ) l_addi( r2, r2, fc );
	}
	else {
		if ( fc >> 16 ) {
			o( (6 << 26) | (r2 << 21) | ((fc >> 16) & 0xffff) ); // l.movhi r2, const_hi
			l_ori( r2, r2, fc );                                 // l.ori   r2, r2, const_lo
		}
		else
			l_ori( r2, 0, fc );                                  // l.ori r2, r0, const_lo
	}

	if ( size == 1 )
		o( (0x36 << 26) | (r2 << 16) | (r << 11) ); // l.sb 0(r2), r
	else if ( size == 2 )
		o( (0x37 << 26) | (r2 << 16) | (r << 11) ); // l.sh 0(r2), r
	else if ( size == 4 )
		o( (0x35 << 26) | (r2 << 16) | (r << 11) ); // l.sw 0(r2), r
	else if ( size == 8 )
		tcc_error("store long long not supported");
  }
  else if ((v->r & VT_VALMASK) == VT_LOCAL) {
	if ( size == 1 )
		o( (0x36 << 26) | (2 << 16) | (r << 11) | ((fc & 0xf800) << 10) | (fc & 0x7ff) ); // l.sb fc(fp), r
	else if ( size == 2 )
		o( (0x37 << 26) | (2 << 16) | (r << 11) | ((fc & 0xf800) << 10) | (fc & 0x7ff) ); // l.sh fc(fp), r
	else if ( size == 4 )
		o( (0x35 << 26) | (2 << 16) | (r << 11) | ((fc & 0xf800) << 10) | (fc & 0x7ff) ); // l.sw fc(fp), r
	else if ( size == 8 )
		tcc_error("store long long not supported");
  }
  else {
	if ( size == 1 )
		o( (0x36 << 26) | (fr << 16) | (r << 11) ); // l.sb 0(fr), r
	else if ( size == 2 )
		o( (0x37 << 26) | (fr << 16) | (r << 11) ); // l.sh 0(fr), r
	else if ( size == 4 )
		o( (0x35 << 26) | (fr << 16) | (r << 11) ); // l.sw 0(fr), r
	else if ( size == 8 )
		tcc_error("store long long not supported");
  }
}

int ld2val_type( int val_type, int ld_type ) {
  if ( ld_type & VT_LVAL_BYTE ) val_type = VT_BYTE;
  else if ( ld_type & VT_LVAL_SHORT ) val_type = VT_SHORT;

  if ( ld_type & VT_LVAL_UNSIGNED ) val_type |= VT_UNSIGNED;

  return val_type;
}

ST_FUNC void gfunc_call(int nb_args) {
  int i, v, r, reg = nb_args, max_regs = 6;
  int arg64_reg_nb = 0, arg64_stack_nb = 0;
  int arg_space = 0;
  int arg_offs = 0;
  Sym *func_sym, *sym_arg;

// commit pending op
  if ( nb_args > 0 ) {
	v = vtop->r & VT_VALMASK;
        if (v == VT_CMP || (v & ~1) == VT_JMP)
            gv(RC_INT);
  }

  func_sym = vtop[-nb_args].type.ref;
  if ( func_sym->f.func_type == FUNC_ELLIPSIS ) {
  	max_regs = 0;
  	sym_arg = func_sym->next;
	while ( sym_arg && max_regs < 6 ) {
		max_regs++;
		sym_arg = sym_arg->next;
	}
  }

  for (i = 0; i < nb_args; i++ ) {
	SValue *sv = &vtop[-i];
	int align;
	int bt = (sv->type.t & VT_BTYPE), size = type_size(&sv->type, &align);
	if ( bt == VT_STRUCT ) {
// save struct args on stack and pass a pointer
		size = (size + 3) & -4;

		r = get_reg(RC_INT);
		l_addi( r, 2, -arg_space - size + loc );
		l_addi( 1, 1, -size );

		vset( &sv->type, r | VT_LVAL, 0 );
		vpushv( sv );
		vstore();
		vtop--;

		arg_space += size;

		sv->type.ref = NULL;
		sv->type.t =  VT_INT;
		sv->r &= ~VT_LVAL;
		sv->r = (sv->r & ~VT_VALMASK) | VT_LOCAL;
		sv->c.i = -arg_space + loc;
	}
  }

  for (i = nb_args-1; i >= 0; i--) {
	SValue *sv = &vtop[-i];
	int bt = (sv->type.t & VT_BTYPE);

	if ( bt == VT_LLONG || fpdbl_type(bt) == VT_DOUBLE ) {
		if ( nb_args - i < max_regs ) {
			arg64_reg_nb++;
			max_regs--;
		}
		else {
			// only one reg left, it will be unused
			if ( nb_args - i == max_regs ) max_regs--;
			arg64_stack_nb++;
		}
	}
  }

  if ( nb_args > max_regs ) {
  	int param_space = (nb_args - max_regs + arg64_stack_nb) * 4;
  	l_addi( 1, 1, -param_space );
  	arg_space += param_space;
	reg = max_regs;
	arg_offs = param_space - 4;
  }

// two regs for int64
  reg += arg64_reg_nb;

  for (i = 0; i < nb_args; i++) {
	int bt = vtop->type.t & VT_BTYPE;

	if ( i < nb_args - max_regs ) {
  		r = gv(RC_INT);
		if ( bt == VT_LLONG || fpdbl_type(bt) == VT_DOUBLE ) {
			g(0xd4); // l.sw i*4(r1), r
			g(0x01);
			v = (vtop->r2 << 11) | arg_offs;
			g(v >> 8);
			g(v);
			arg_offs -= 4;
		}

		g(0xd4); // l.sw i*4(r1), r
		g(0x01);
		v = (r << 11) | arg_offs;
		g(v >> 8);
		g(v);
		arg_offs -= 4;
	}
	else {
		if ( bt == VT_LLONG || fpdbl_type(bt) == VT_DOUBLE ) {
			r = gv(RC_INT);
			l_ori( reg + 2, vtop->r2, 0 );
			reg--;
			l_ori( reg + 2, r, 0 );
			reg--;
		}
		else {
			save_reg( reg+2 );
			vtop->type.t = ld2val_type( vtop->type.t, vtop->r );
			load( reg+2, vtop );
			reg--;
		}
	}

	vtop--;
  }

  save_regs(0);

  if ((vtop->r & (VT_VALMASK | VT_LVAL)) == VT_CONST && (vtop->r & VT_SYM)) {
	greloc(cur_text_section, vtop->sym, ind, R_OR1K_INSN_REL_26);
	o( (1 << 26) | vtop->c.i ); // l.jal rel26
  }
  else {
  	int r = gv(RC_INT);
	o( (0x12 << 26) | (r << 11) ); // l.jalr r
  }

  o(0x15000000);              // l.nop

  if ( arg_space ) l_addi( 1, 1, arg_space );

  vtop--;
}

ST_FUNC void gfunc_prolog(CType *func_type) {
  int bt, addr, /* align,*/ reg = 0, max_reg = 6, i;
  Sym *sym;
  CType *type;

  sym = func_type->ref;
  func_vt = sym->type;

  loc = 0;
  addr = 8;

  if ( (func_vt.t & VT_BTYPE) == VT_STRUCT ) {
  	func_vc = addr;
  	addr += 4;
  	reg++;
	max_reg++;
  }

  for ( sym = func_type->ref->next; sym; sym = sym->next ) {
	type = &sym->type;
	bt   = type->t & VT_BTYPE;

	if ( bt == VT_STRUCT )
		sym_push(sym->v & ~SYM_FIELD, type, VT_LLOCAL | lvalue_type(type->t), addr);
	else {
		int offs = 0;
		if ( bt == VT_BYTE ) offs = 3;
		else if ( bt == VT_SHORT ) offs = 2;

		sym_push(sym->v & ~SYM_FIELD, type, VT_LOCAL  | lvalue_type(type->t), addr + offs);
	}

	addr += 4;

	if ( bt == VT_LLONG || fpdbl_type(bt) == VT_DOUBLE ) {
		addr += 4;
		if ( reg+1 < max_reg ) reg += 2; else max_reg = 0;
	}
	else {
	  	if ( reg < max_reg ) reg++;
	}
  }

  if ( reg ) l_addi( 1, 1, -reg * 4);
  func_ret_sub = reg * 4;

  for ( i = 0; i < reg; i++ )
	o((0x35 << 26) | (1 << 16) | ((i+3) << 11) | ((i*4) & 0x7ff) ); // l.sw offs(r1), r

  l_addi( 1, 1, -8 );
  o(0xd4014804); // l.sw -4(r1),r9
  o(0xd4011000); // l.sw -8(r1), r2
  l_ori ( 2, 1,  0 );

  func_sub_sp_offset = ind;
  l_addi( 1, 1, 0 ); // protect local vars, will be patched in gfunc_epilog
}

ST_FUNC void gfunc_epilog(void) {
  int local = (-loc + 3) & -4;
  cur_text_section->data[func_sub_sp_offset+2] = (-local) >> 8;
  cur_text_section->data[func_sub_sp_offset+3] = -local;

  l_ori( 1, 2, 0 );
  o(0x84410000); // l.lwz  r2, 0(r1)
  o(0x85210004); // l.lwz  r9, 4(r1)
  l_addi( 1, 1, func_ret_sub + 8 );
  o(0x44004800); // l.jr   r9
  o(0x15000000); // l.nop
}

ST_FUNC void gen_fill_nops(int bytes) {
  if ((bytes & 3))
	tcc_error("alignment of code section not multiple of 4");

  while (bytes > 0) {
	o(0x15000000); // l.nop
	bytes -= 4;
  }
}

void set_top_type( int bt ) {
  vtop->type.t &= ~VT_TYPE;
  vtop->type.t |= bt;
}

void gcall_int_func( int ft, int func ) {
  vpush_global_sym(&func_old_type, func);
  vrott(3);
  gfunc_call(2);
  vpushi(0);
  set_top_type( ft );
  vtop->r = REG_IRET;
}

#define get_opregs()     \
  gv2( RC_INT, RC_INT ); \
  r  = vtop[-1].r;       \
  fr = vtop[ 0].r;       \
  vtop--;

#define gen_alu_op(op_code) \
  get_opregs()              \
  o( (0x38 << 26) | (r << 21) | (r << 16) | (fr << 11)| op_code );

#define gen_alu_op_imm(op_code, imm_op_code) \
  if ( (vtop->r & (VT_VALMASK | VT_LVAL | VT_SYM)) == VT_CONST && vtop->c.i <= 0x7fff ) { \
  	fc = vtop->c.i; \
	vtop--; \
	r = gv( RC_INT ); \
        o( (imm_op_code << 26) | (r << 21) | (r << 16) | fc ); \
  } \
  else { \
  	gen_alu_op(op_code) \
  }

#define gen_fl_op(op_code, imm_op_code) \
  if ( (vtop->r & (VT_VALMASK | VT_LVAL | VT_SYM)) == VT_CONST && vtop->c.i <= 0x7fff ) { \
  	fc = vtop->c.i; \
	vtop--; \
	r = gv( RC_INT ); \
  	o( (imm_op_code << 21) | (r << 16) | fc ); \
  } \
  else { \
	get_opregs() \
  	o( (op_code << 21) | (r << 16) | (fr << 11) ); \
  } \
  vset_VT_CMP(0x80);

ST_FUNC void gen_opi(int op) {
  int r, fr, fc;

  switch (op) {
  case'+':
  case TOK_ADDC1:
 	gen_alu_op_imm( 0, 0x27 ); break; // l.add/addi r, r, fr/fc

  case TOK_ADDC2:
 	gen_alu_op_imm( 1, 0x28 ); break; // l.addc/addic r, r, fr/fc

  case '-':
  case TOK_SUBC1:
	gen_alu_op( 2 ); break; // l.sub r, r, fr

  case '&': gen_alu_op_imm( 3, 0x29 ); break; // l.and r, r, fr
  case '^': gen_alu_op_imm( 5, 0x2b ); break; // l.xor r, r, fr
  case '|': gen_alu_op_imm( 4, 0x2a ); break; // l.or r, r, fr
  case TOK_SHL: gen_alu_op( 8 ); break; // l.sll r, r, fr
  case TOK_SHR: gen_alu_op( 0x48 ); break; // l.srl r, r, fr
  case TOK_SAR: gen_alu_op( 0x88 ); break; // l.sra r, r, fr

  case '*':
 	gen_alu_op( 0x306 ); break; // l.mul r, r, fr

  case TOK_UMULL: {
  	CType ll_type;
	ll_type.t = VT_LLONG | VT_UNSIGNED;
	vpush_global_sym(&func_old_type, TOK___muldi3);
	vrott(3);
	gen_cast(&ll_type);
	vswap();
	gen_cast(&ll_type);
	gfunc_call(2);
	vpushi(0);
	set_top_type( VT_LLONG | VT_UNSIGNED );
	vtop->r  = REG_IRET;
	vtop->r2 = REG_LRET;
	break;
    }

  case '/':
  case TOK_PDIV:
 	gen_alu_op( 0x309 ); break; // l.div r, r, fr

  case TOK_UDIV:
 	gen_alu_op( 0x30a ); break; // l.divu r, r, fr

  case TOK_EQ: gen_fl_op( 0x720, 0x5e0 ); break; // l.sfeq r, fr
  case TOK_NE: gen_fl_op( 0x721, 0x5e1 ); break; // l.sfne r, fr
  case TOK_LT: gen_fl_op( 0x72c, 0x5ec ); break; // l.sflts r, fr
  case TOK_GE: gen_fl_op( 0x72b, 0x5eb ); break; // l.sfges r, fr
  case TOK_GT: gen_fl_op( 0x72a, 0x5ea ); break; // l.sfgts r, fr
  case TOK_LE: gen_fl_op( 0x72d, 0x5ed ); break; // l.sfles r, fr
  case TOK_ULT: gen_fl_op( 0x724, 0x5e4 ); break; // l.sfltu r, fr
  case TOK_UGE: gen_fl_op( 0x723, 0x5e3 ); break; // l.sfgeu r, fr
  case TOK_UGT: gen_fl_op( 0x722, 0x5e2 ); break; // l.sfgtu r, fr
  case TOK_ULE: gen_fl_op( 0x725, 0x5e5 ); break; // l.sfleu r, fr

  case '%':      gcall_int_func( VT_INT,             TOK___modsi3 );  break;
  case TOK_UMOD: gcall_int_func( VT_INT|VT_UNSIGNED, TOK___umodsi3 ); break;

  default:
	tcc_error( "unsupported tok->op 0x%0x", op );
  }
}

void gcall_fp_func( int ft, int func, int func_dbl ) {
  vpush_global_sym(&func_old_type, ((ft & VT_BTYPE) == VT_FLOAT)? func: func_dbl);
  vrott(3);
  gfunc_call(2);
  vpushi(0);
  set_top_type( ft );
  vtop->r  = REG_IRET;
  vtop->r2 = ((ft & VT_BTYPE) == VT_FLOAT)? VT_CONST: REG_LRET;
}

void gcall_fpcmp_func( int op, int ft, int func, int func_dbl ) {
  vpush_global_sym(&func_old_type, ((ft & VT_BTYPE) == VT_FLOAT)? func: func_dbl);
  vrott(3);
  gfunc_call(2);
  vpushi(0);
  set_top_type( VT_INT );
  vtop->r = REG_IRET;
  vpushi(0);
  gen_opi(op);
}

ST_FUNC void gen_opf(int op) {
  int ft = vtop->type.t & VT_BTYPE;

  switch (op) {
  case '+': gcall_fp_func( ft, TOK___addsf3, TOK___adddf3 ); break;
  case '-': gcall_fp_func( ft, TOK___subsf3, TOK___subdf3 ); break;
  case '*': gcall_fp_func( ft, TOK___mulsf3, TOK___muldf3 ); break;
  case '/': gcall_fp_func( ft, TOK___divsf3, TOK___divdf3 ); break;

#ifndef CMPX_FLT
  case TOK_EQ: gcall_fpcmp_func( op, ft, TOK___eqsf2, TOK___eqdf2 ); break;
  case TOK_NE: gcall_fpcmp_func( op, ft, TOK___nesf2, TOK___nedf2 ); break;
  case TOK_LT: gcall_fpcmp_func( op, ft, TOK___ltsf2, TOK___ltdf2 ); break;
  case TOK_GE: gcall_fpcmp_func( op, ft, TOK___gesf2, TOK___gedf2 ); break;
  case TOK_GT: gcall_fpcmp_func( op, ft, TOK___gtsf2, TOK___gtdf2 ); break;
  case TOK_LE: gcall_fpcmp_func( op, ft, TOK___lesf2, TOK___ledf2 ); break;
#else
  	if (op >= TOK_EQ && op <= TOK_GT) {
		gcall_fp_func( ft, TOK___cmpsf2, TOK___cmpdf2 );
  		set_top_type( VT_INT );
		vpushi(0);
		gen_opi(op);
		break;
	}
#endif

	tcc_error( "unsupported floating tok->op 0x%0x", op );
  }
}

ST_FUNC void gen_cvt_itof(int t) {
  int bt1    = t & VT_BTYPE,
      bt2    = vtop->type.t & VT_BTYPE,
      unsign = vtop->type.t & VT_UNSIGNED;

  if ( bt2 == VT_INT ) {
	if ( unsign )
		vpush_global_sym(&func_old_type, (bt1 == VT_FLOAT)? TOK___floatunsisf: TOK___floatunsidf);
	else
		vpush_global_sym(&func_old_type, (bt1 == VT_FLOAT)? TOK___floatsisf: TOK___floatsidf);
  }
  else if ( bt2 == VT_LLONG ) {
	if ( unsign )
		vpush_global_sym(&func_old_type, (bt1 == VT_FLOAT)? TOK___floatundisf: TOK___floatundidf);
	else
		vpush_global_sym(&func_old_type, (bt1 == VT_FLOAT)? TOK___floatdisf: TOK___floatdidf);
  }
  else
	tcc_error( "unsupported type 0x%x for int2float conversion", bt2 );

  vswap();
  gfunc_call(1);
  vpushi(0);
  set_top_type( bt1 | unsign );
  vtop->r  = REG_IRET;
  vtop->r2 = ((bt1 & VT_BTYPE) == VT_FLOAT)? VT_CONST: REG_LRET;
}

ST_FUNC void gen_cvt_ftoi(int t) {
  int bt = vtop->type.t & VT_BTYPE;

  if ( (t & VT_BTYPE) == VT_INT ) {
	if ( t & VT_UNSIGNED )
		vpush_global_sym(&func_old_type, (bt == VT_FLOAT)? TOK___fixunssfsi: TOK___fixunsdfsi);
  	else
		vpush_global_sym(&func_old_type, (bt == VT_FLOAT)? TOK___fixsfsi: TOK___fixdfsi);
  }
  else if ( (t & VT_BTYPE) == VT_LLONG ) {
	if ( t & VT_UNSIGNED )
		vpush_global_sym(&func_old_type, (bt == VT_FLOAT)? TOK___fixunssfdi: TOK___fixunsdfdi);
  	else
		vpush_global_sym(&func_old_type, (bt == VT_FLOAT)? TOK___fixsfdi: TOK___fixdfdi);
  }
  else
	tcc_error( "unsupported type 0x%x for float2int conversion", t );

  vswap();
  gfunc_call(1);
  vpushi(0);
  set_top_type( bt | (t & VT_UNSIGNED) );
  vtop->r  = REG_IRET;
  vtop->r2 = (bt == VT_INT)? VT_CONST: REG_LRET;
}

ST_FUNC void gen_cvt_ftof(int t) {
  int bt1 = vtop->type.t & VT_BTYPE,
      bt2 = t & VT_BTYPE;

  if ( bt1 == VT_FLOAT && fpdbl_type(bt2) == VT_DOUBLE )
  		vpush_global_sym(&func_old_type, TOK___extendsfdf2);
  else if ( fpdbl_type(bt1) == VT_DOUBLE && bt2 == VT_FLOAT )
  		vpush_global_sym(&func_old_type, TOK___truncdfsf2);
// double to long double and vice versa
  else if ( fpdbl_type(bt1) == VT_DOUBLE && fpdbl_type(bt2) == VT_DOUBLE ) {
  	if ( ((vtop->r & VT_VALMASK) == VT_CONST) && bt2 == VT_LDOUBLE ) vtop->c.ld = vtop->c.d;
	else if ( ((vtop->r & VT_VALMASK) == VT_CONST) && bt2 == VT_DOUBLE ) vtop->c.d = vtop->c.ld;
  	set_top_type( bt2 );
	return;
  }
  else
  	tcc_error( "unsupported float2float conversion" );

  vswap();
  gfunc_call(1);
  vpushi(0);
  set_top_type( bt2 );
  vtop->r  = REG_IRET;
  vtop->r2 = (bt2 == VT_FLOAT)? VT_CONST: REG_LRET;
}

ST_FUNC int gjmp_cond(int op, int t) {
  int ind1 = ind;
  int inv = op & 1;

  if ( nocode_wanted ) return t;

  if ( inv )
	o( rel26enc(3, t - ind1) ); // l.bnf t
  else
	o( rel26enc(4, t - ind1) ); // l.bf t
  o(0x15000000); // l.nop

  return ind1;
}

ST_FUNC int gjmp_append(int n0, int t) {
  int n = n0;
  unsigned char *ptr = 0;
  uint32_t mval = 0;
  int32_t disp;

  if (!n0) return t;

  while ( n ) {
	ptr = cur_text_section->data + n;
	mval = read32be( ptr );
	disp = rel26dec( mval );
	if ( !disp || n + disp * 4 == 0 ) break;

	n += disp * 4;
  }
  write32be( ptr, rel26enc(mval >> 26, t - n) );

  return n0;
}

ST_FUNC void ggoto(void) {
  tcc_error( "ggoto not implemented" );
}

/* number of registers for returning a struct, 0 means return is thru a pointer */
ST_FUNC int gfunc_sret(CType *vt, int variadic, CType *ret, int *ret_align, int *regsize) {
  return 0;
}

ST_FUNC void gen_vla_sp_save(int addr) {
  tcc_error("variable length arrays unsupported for this target");
}

ST_FUNC void gen_vla_sp_restore(int addr) {
  tcc_error("variable length arrays unsupported for this target");
}

ST_FUNC void gen_vla_alloc(CType *type, int align) {
}

#endif // TARGET_DEFS_ONLY

