#ifdef TARGET_DEFS_ONLY

#define EM_TCC_TARGET EM_OPENRISC

#define R_DATA_32   R_OR1K_32
#define R_DATA_PTR  R_OR1K_32
#define R_JMP_SLOT  R_OR1K_JMP_SLOT
#define R_GLOB_DAT  R_OR1K_GLOB_DAT
#define R_COPY      R_OR1K_COPY
#define R_RELATIVE  R_OR1K_RELATIVE

#define ELF_START_ADDR 0x2000
#define ELF_PAGE_SIZE  0x2000

#define PCRELATIVE_DLLPLT 0
#define RELOCATE_DLLPLT 0

#else // TARGET_DEFS_ONLY

#include "tcc.h"

int code_reloc(int reloc_type) {
  switch (reloc_type) {
  case R_OR1K_32:
  case R_OR1K_RELATIVE:
  case R_OR1K_32_PCREL:
  case R_OR1K_COPY:
  case R_OR1K_GLOB_DAT:
  case R_OR1K_HI_16_IN_INSN:
  case R_OR1K_LO_16_IN_INSN:
  case R_OR1K_NONE:
  	return 0;

  case R_OR1K_JMP_SLOT:
  case R_OR1K_INSN_REL_26:
  	return 1;
  }

  tcc_error("Unknown relocation type: %d", reloc_type);
  return -1;
}

int gotplt_entry_type(int reloc_type) {
  switch (reloc_type) {
  case R_OR1K_COPY:
  case R_OR1K_GLOB_DAT:
  case R_OR1K_JMP_SLOT:
  case R_OR1K_NONE:
  	return NO_GOTPLT_ENTRY;

  case R_OR1K_HI_16_IN_INSN:
  case R_OR1K_LO_16_IN_INSN:
  case R_OR1K_INSN_REL_26:
  	return AUTO_GOTPLT_ENTRY;

  case R_OR1K_32:
  case R_OR1K_RELATIVE:
  case R_OR1K_32_PCREL:

  case R_OR1K_GOTPC_HI16:
  case R_OR1K_GOTPC_LO16:
  case R_OR1K_GOT16:
  case R_OR1K_PLT26:
  case R_OR1K_GOTOFF_HI16:
  case R_OR1K_GOTOFF_LO16:
	return AUTO_GOTPLT_ENTRY;
  }

  tcc_error("Unknown relocation type: %d", reloc_type);
  return -1;
}

ST_FUNC unsigned create_plt_entry(TCCState *s1, unsigned got_offset, struct sym_attr *attr) {
  Section *plt = s1->plt;
  uint32_t plt_offset;
  unsigned char *p;

  if ( plt->data_offset == 0 ) {
  	p = section_ptr_add( plt, 20 );
	write32be( p,    0x19800000 ); // l.movhi r12, hi(got)
	write32be( p+4,  0xa98c0000 ); // l.ori r12, r12, lo(got)
	write32be( p+8,  0x85ec0004 ); // l.lwz r15, 4(r12)
	write32be( p+12, 0x44007800 ); // l.jr r15
	write32be( p+16, 0x858c0000 ); // l.lwz r12, 0(r12)
  }

  plt_offset = plt->data_offset;
  p = section_ptr_add( plt, 20 );

  p[0] = 0x19; // l.movhi r12, hi(got_offset)
  p[1] = 0x80;
  p[2] = got_offset >> 24;
  p[3] = got_offset >> 16;
  p[4] = 0xa9; // l.ori r12, r12, lo(got_offset)(r12)
  p[5] = 0x8c;
  p[6] = got_offset >> 8;
  p[7] = got_offset;

  write32be( p+8,  0x858c0000 ); // l.lwz r12, 0(r12)
  write32be( p+12, 0x44006000 ); // l.jr r12
  write32be( p+16, 0xa9600000 ); // l.ori r11, r0, 0x0

  return plt_offset;
}

ST_FUNC void relocate_plt(TCCState *s1) {
  unsigned char *p, *p_end;
  uint32_t v32;

  if ( !s1->plt ) return;

  p     = s1->plt->data;
  p_end = p + s1->plt->data_offset;

  while ( p < p_end ) {
  	v32 = (p[2] << 24) | (p[3] << 16) | (p[6] << 8) | p[7];
	v32 += s1->got->sh_addr;

	p[2] = v32 >> 24;
	p[3] = v32 >> 16;
	p[6] = v32 >> 8;
	p[7] = v32;
	
  	p += 20;
  }
}

void relocate_init(Section *sr) {
}

void relocate(TCCState *s1, ElfW_Rel *rel, int type, unsigned char *ptr, addr_t addr, addr_t val) {
  uint32_t code;
  int sym_index = ELFW(R_SYM)(rel->r_info);

//  printf("link: relocate at 0x%x\n", addr);

  switch (type) {
  case R_OR1K_PLT26:
  case R_OR1K_INSN_REL_26:
  	code = read32be(ptr);
	write32be( ptr, (code & (0x3f << 26)) | ((((code & 0x3ffffff) + val - addr) >> 2) & 0x3ffffff) );
  	break;

  case R_OR1K_LO_16_IN_INSN:
	code = read32be(ptr);
	write32be( ptr, (code & 0xffff0000) | (val & 0xffff) );
	break;

  case R_OR1K_HI_16_IN_INSN:
	code = read32be(ptr);
	write32be( ptr, (code & 0xffff0000) | (val >> 16) );
  	break;

  case R_OR1K_32_PCREL:
  	code = read32be(ptr);
	write32be( ptr, code + val - addr );
	break;

  case R_OR1K_32:
  	code = read32be(ptr);
	write32be( ptr, code + val );
	break;

#if 0
  case R_OR1K_COPY:
	write32be( ptr, val );
	break;
#endif

  case R_OR1K_GOTPC_HI16:
  	code = read32be(ptr);
  	write32le( ptr, (code & 0xffff0000) | ((code + s1->got->sh_addr - addr) >> 16) );
  	break;

  case R_OR1K_GOTPC_LO16:
  	code = read32be(ptr);
  	write32le( ptr, (code & 0xffff0000) | ((code + s1->got->sh_addr - addr) & 0xffff) );
  	break;

  case R_OR1K_GOT16:
  	code = read32be(ptr);
  	write32le( ptr, (code & 0xffff0000) | ((code + s1->sym_attrs[sym_index].got_offset) & 0xffff) );
	break;

  case R_OR1K_GOTOFF_HI16:
  	code = read32be(ptr);
  	write32le( ptr, (code & 0xffff0000) | ((code + val - s1->got->sh_addr) >> 16) );
	break;

  case R_OR1K_GOTOFF_LO16:
  	code = read32be(ptr);
  	write32le( ptr, (code & 0xffff0000) | ((code + val - s1->got->sh_addr) & 0xffff) );
  	break;

  default:
  	tcc_error("unsupported reloc type 0x%x", type);
  }
}

#endif // TARGET_DEFS_ONLY
